/*
 * Copyright 2019-2021 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.springframework.data.elasticsearch.core.convert;

import static java.util.Collections.*;
import static org.assertj.core.api.Assertions.*;
import static org.skyscreamer.jsonassert.JSONAssert.*;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.OffsetTime;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.json.JSONException;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.core.convert.ConversionService;
import org.springframework.core.convert.converter.Converter;
import org.springframework.core.convert.support.GenericConversionService;
import org.springframework.data.annotation.Id;
import org.springframework.data.annotation.ReadOnlyProperty;
import org.springframework.data.annotation.Transient;
import org.springframework.data.annotation.TypeAlias;
import org.springframework.data.convert.ReadingConverter;
import org.springframework.data.convert.WritingConverter;
import org.springframework.data.elasticsearch.annotations.DateFormat;
import org.springframework.data.elasticsearch.annotations.Field;
import org.springframework.data.elasticsearch.annotations.FieldType;
import org.springframework.data.elasticsearch.annotations.GeoPointField;
import org.springframework.data.elasticsearch.core.Range;
import org.springframework.data.elasticsearch.core.document.Document;
import org.springframework.data.elasticsearch.core.geo.GeoJsonEntity;
import org.springframework.data.elasticsearch.core.geo.GeoJsonGeometryCollection;
import org.springframework.data.elasticsearch.core.geo.GeoJsonLineString;
import org.springframework.data.elasticsearch.core.geo.GeoJsonMultiLineString;
import org.springframework.data.elasticsearch.core.geo.GeoJsonMultiPoint;
import org.springframework.data.elasticsearch.core.geo.GeoJsonMultiPolygon;
import org.springframework.data.elasticsearch.core.geo.GeoJsonPoint;
import org.springframework.data.elasticsearch.core.geo.GeoJsonPolygon;
import org.springframework.data.elasticsearch.core.geo.GeoPoint;
import org.springframework.data.elasticsearch.core.mapping.SimpleElasticsearchMappingContext;
import org.springframework.data.elasticsearch.core.query.SeqNoPrimaryTerm;
import org.springframework.data.geo.Box;
import org.springframework.data.geo.Circle;
import org.springframework.data.geo.Point;
import org.springframework.data.geo.Polygon;
import org.springframework.data.mapping.context.MappingContext;
import org.springframework.lang.Nullable;

/**
 * Unit tests for {@link MappingElasticsearchConverter}.
 *
 * @author Christoph Strobl
 * @author Mark Paluch
 * @author Peter-Josef Meisch
 * @author Konrad Kurdej
 * @author Roman Puchkovskiy
 * @author Sascha Woo
 */
public class MappingElasticsearchConverterUnitTests {

	static final String JSON_STRING = "{\"_class\":\"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$Car\",\"name\":\"Grat\",\"model\":\"Ford\"}";
	static final String CAR_MODEL = "Ford";
	static final String CAR_NAME = "Grat";
	MappingElasticsearchConverter mappingElasticsearchConverter;

	Person sarahConnor;
	Person kyleReese;
	Person t800;

	Inventory gun = new Gun("Glock 19", 33);
	Inventory grenade = new Grenade("40 mm");
	Inventory rifle = new Rifle("AR-18 Assault Rifle", 3.17, 40);
	Inventory shotGun = new ShotGun("Ithaca 37 Pump Shotgun");

	Address observatoryRoad;
	Place bigBunsCafe;

	Document sarahAsMap;
	Document t800AsMap;
	Document kyleAsMap;
	Document gratiotAveAsMap;
	Document locationAsMap;
	Document gunAsMap;
	Document grenadeAsMap;
	Document rifleAsMap;
	Document shotGunAsMap;
	Document bigBunsCafeAsMap;
	Document notificationAsMap;

	@BeforeEach
	public void init() {

		SimpleElasticsearchMappingContext mappingContext = new SimpleElasticsearchMappingContext();
		mappingContext.setInitialEntitySet(Collections.singleton(Rifle.class));
		mappingContext.afterPropertiesSet();

		mappingElasticsearchConverter = new MappingElasticsearchConverter(mappingContext, new GenericConversionService());
		mappingElasticsearchConverter.setConversions(
				new ElasticsearchCustomConversions(Arrays.asList(new ShotGunToMapConverter(), new MapToShotGunConverter())));
		mappingElasticsearchConverter.afterPropertiesSet();

		sarahConnor = new Person();
		sarahConnor.id = "sarah";
		sarahConnor.name = "Sarah Connor";
		sarahConnor.gender = Gender.MAN;

		kyleReese = new Person();
		kyleReese.id = "kyle";
		kyleReese.gender = Gender.MAN;
		kyleReese.name = "Kyle Reese";

		t800 = new Person();
		t800.id = "t800";
		t800.name = "T-800";
		t800.gender = Gender.MACHINE;

		t800AsMap = Document.create();
		t800AsMap.put("id", "t800");
		t800AsMap.put("name", "T-800");
		t800AsMap.put("gender", "MACHINE");
		t800AsMap.put("_class",
				"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$Person");

		observatoryRoad = new Address();
		observatoryRoad.city = "Los Angeles";
		observatoryRoad.street = "2800 East Observatory Road";
		observatoryRoad.location = new Point(-118.3026284D, 34.118347D);

		bigBunsCafe = new Place();
		bigBunsCafe.setName("Big Buns Cafe");
		bigBunsCafe.setCity("Los Angeles");
		bigBunsCafe.setStreet("15 South Fremont Avenue");
		bigBunsCafe.setLocation(new Point(-118.1545845D, 34.0945637D));

		sarahAsMap = Document.create();
		sarahAsMap.put("id", "sarah");
		sarahAsMap.put("name", "Sarah Connor");
		sarahAsMap.put("gender", "MAN");
		sarahAsMap.put("_class",
				"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$Person");

		kyleAsMap = Document.create();
		kyleAsMap.put("id", "kyle");
		kyleAsMap.put("gender", "MAN");
		kyleAsMap.put("name", "Kyle Reese");

		locationAsMap = Document.create();
		locationAsMap.put("lat", 34.118347D);
		locationAsMap.put("lon", -118.3026284D);

		gratiotAveAsMap = Document.create();
		gratiotAveAsMap.put("city", "Los Angeles");
		gratiotAveAsMap.put("street", "2800 East Observatory Road");
		gratiotAveAsMap.put("location", locationAsMap);

		bigBunsCafeAsMap = Document.create();
		bigBunsCafeAsMap.put("name", "Big Buns Cafe");
		bigBunsCafeAsMap.put("city", "Los Angeles");
		bigBunsCafeAsMap.put("street", "15 South Fremont Avenue");
		bigBunsCafeAsMap.put("location", new LinkedHashMap<>());
		((HashMap<String, Object>) bigBunsCafeAsMap.get("location")).put("lat", 34.0945637D);
		((HashMap<String, Object>) bigBunsCafeAsMap.get("location")).put("lon", -118.1545845D);
		bigBunsCafeAsMap.put("_class",
				"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$Place");

		gunAsMap = Document.create();
		gunAsMap.put("label", "Glock 19");
		gunAsMap.put("shotsPerMagazine", 33);
		gunAsMap.put("_class", Gun.class.getName());

		grenadeAsMap = Document.create();
		grenadeAsMap.put("label", "40 mm");
		grenadeAsMap.put("_class", Grenade.class.getName());

		rifleAsMap = Document.create();
		rifleAsMap.put("label", "AR-18 Assault Rifle");
		rifleAsMap.put("weight", 3.17D);
		rifleAsMap.put("maxShotsPerMagazine", 40);
		rifleAsMap.put("_class", "rifle");

		shotGunAsMap = Document.create();
		shotGunAsMap.put("model", "Ithaca 37 Pump Shotgun");
		shotGunAsMap.put("_class", ShotGun.class.getName());

		notificationAsMap = Document.create();
		notificationAsMap.put("id", 1L);
		notificationAsMap.put("fromEmail", "from@email.com");
		notificationAsMap.put("toEmail", "to@email.com");
		Map<String, Object> data = new HashMap<>();
		data.put("documentType", "abc");
		data.put("content", null);
		notificationAsMap.put("params", data);
		notificationAsMap.put("_class",
				"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$Notification");
	}

	@Test
	public void shouldFailToInitializeGivenMappingContextIsNull() {

		// given
		assertThatThrownBy(() -> new MappingElasticsearchConverter(null)).isInstanceOf(IllegalArgumentException.class);
	}

	@Test
	public void shouldReturnMappingContextWithWhichItWasInitialized() {

		// given
		MappingContext mappingContext = new SimpleElasticsearchMappingContext();
		MappingElasticsearchConverter converter = new MappingElasticsearchConverter(mappingContext);

		// then
		assertThat(converter.getMappingContext()).isNotNull();
		assertThat(converter.getMappingContext()).isSameAs(mappingContext);
	}

	@Test
	public void shouldReturnDefaultConversionService() {

		// given
		MappingElasticsearchConverter converter = new MappingElasticsearchConverter(
				new SimpleElasticsearchMappingContext());

		// when
		ConversionService conversionService = converter.getConversionService();

		// then
		assertThat(conversionService).isNotNull();
	}

	@Test // DATAES-530
	public void shouldMapObjectToJsonString() {
		Car car = new Car();
		car.setModel(CAR_MODEL);
		car.setName(CAR_NAME);
		String jsonResult = mappingElasticsearchConverter.mapObject(car).toJson();

		assertThat(jsonResult).isEqualTo(JSON_STRING);
	}

	@Test // DATAES-530
	public void shouldReadJsonStringToObject() {
		// Given

		// When
		Car result = mappingElasticsearchConverter.read(Car.class, Document.parse(JSON_STRING));

		// Then
		assertThat(result).isNotNull();
		assertThat(result.getName()).isEqualTo(CAR_NAME);
		assertThat(result.getModel()).isEqualTo(CAR_MODEL);
	}

	@Test // DATAES-530
	public void shouldMapGeoPointElasticsearchNames() throws JSONException {
		// given
		double lon = 5;
		double lat = 48;
		Point point = new Point(lon, lat);
		// ES has Strings in "lat,lon", but has arrays as [lon,lat]!!
		String pointAsString = lat + "," + lon;
		double[] pointAsArray = { lon, lat };

		String expected = "{\n" + //
				"  \"pointA\": {\n" + //
				"    \"lon\": 5.0,\n" + //
				"    \"lat\": 48.0\n" + //
				"  },\n" + //
				"  \"pointB\": {\n" + //
				"    \"lon\": 5.0,\n" + //
				"    \"lat\": 48.0\n" + //
				"  },\n" + //
				"  \"pointC\": \"48.0,5.0\",\n" + //
				"  \"pointD\": [\n" + //
				"    5.0,\n" + //
				"    48.0\n" + //
				"  ]\n" + //
				"}\n"; //

		GeoEntity geoEntity = new GeoEntity();
		geoEntity.setPointA(point);
		geoEntity.setPointB(GeoPoint.fromPoint(point));
		geoEntity.setPointC(pointAsString);
		geoEntity.setPointD(pointAsArray);
		String jsonResult = mappingElasticsearchConverter.mapObject(geoEntity).toJson();

		assertEquals(expected, jsonResult, false);
	}

	@Test // DATAES-530
	public void ignoresReadOnlyProperties() {

		// given
		Sample sample = new Sample();
		sample.setReadOnly("readOnly");
		sample.setProperty("property");
		sample.setJavaTransientProperty("javaTransient");
		sample.setAnnotatedTransientProperty("transient");

		// when
		String result = mappingElasticsearchConverter.mapObject(sample).toJson();

		// then
		assertThat(result).contains("\"property\"");
		assertThat(result).contains("\"javaTransient\"");

		assertThat(result).doesNotContain("readOnly");
		assertThat(result).doesNotContain("annotatedTransientProperty");
	}

	@Test // DATAES-530
	public void writesNestedEntity() {

		Person person = new Person();
		person.birthDate = LocalDate.now();
		person.gender = Gender.MAN;
		person.address = observatoryRoad;

		Map<String, Object> sink = writeToMap(person);

		assertThat(sink.get("address")).isEqualTo(gratiotAveAsMap);
	}

	@Test // DATAES-530
	public void writesConcreteList() {

		Person ginger = new Person();
		ginger.setId("ginger");
		ginger.setGender(Gender.MAN);

		sarahConnor.coWorkers = Arrays.asList(kyleReese, ginger);

		Map<String, Object> target = writeToMap(sarahConnor);
		assertThat((List) target.get("coWorkers")).hasSize(2).contains(kyleAsMap);
	}

	@Test // DATAES-530
	public void writesInterfaceList() {

		Inventory gun = new Gun("Glock 19", 33);
		Inventory grenade = new Grenade("40 mm");

		sarahConnor.inventoryList = Arrays.asList(gun, grenade);

		Map<String, Object> target = writeToMap(sarahConnor);
		assertThat((List) target.get("inventoryList")).containsExactly(gunAsMap, grenadeAsMap);
	}

	@Test // DATAES-530
	public void readTypeCorrectly() {

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target).isEqualTo(sarahConnor);
	}

	@Test // DATAES-530
	public void readListOfConcreteTypesCorrectly() {

		sarahAsMap.put("coWorkers", Collections.singletonList(kyleAsMap));

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.getCoWorkers()).contains(kyleReese);
	}

	@Test // DATAES-530
	public void readListOfInterfacesTypesCorrectly() {

		sarahAsMap.put("inventoryList", Arrays.asList(gunAsMap, grenadeAsMap));

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.getInventoryList()).containsExactly(gun, grenade);
	}

	@Test // DATAES-530
	public void writeMapOfConcreteType() {

		sarahConnor.shippingAddresses = new LinkedHashMap<>();
		sarahConnor.shippingAddresses.put("home", observatoryRoad);

		Map<String, Object> target = writeToMap(sarahConnor);
		assertThat(target.get("shippingAddresses")).isInstanceOf(Map.class);
		assertThat(target.get("shippingAddresses")).isEqualTo(Collections.singletonMap("home", gratiotAveAsMap));
	}

	@Test // DATAES-530
	public void writeMapOfInterfaceType() {

		sarahConnor.inventoryMap = new LinkedHashMap<>();
		sarahConnor.inventoryMap.put("glock19", gun);
		sarahConnor.inventoryMap.put("40 mm grenade", grenade);

		Map<String, Object> target = writeToMap(sarahConnor);
		assertThat(target.get("inventoryMap")).isInstanceOf(Map.class);
		assertThat((Map) target.get("inventoryMap")).containsEntry("glock19", gunAsMap).containsEntry("40 mm grenade",
				grenadeAsMap);
	}

	@Test // DATAES-530
	public void readConcreteMapCorrectly() {

		sarahAsMap.put("shippingAddresses", Collections.singletonMap("home", gratiotAveAsMap));

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.getShippingAddresses()).hasSize(1).containsEntry("home", observatoryRoad);
	}

	@Test // DATAES-530
	public void readInterfaceMapCorrectly() {

		sarahAsMap.put("inventoryMap", Collections.singletonMap("glock19", gunAsMap));

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.getInventoryMap()).hasSize(1).containsEntry("glock19", gun);
	}

	@Test // DATAES-530
	public void genericWriteList() {

		Skynet skynet = new Skynet();
		skynet.objectList = new ArrayList<>();
		skynet.objectList.add(t800);
		skynet.objectList.add(gun);

		Map<String, Object> target = writeToMap(skynet);

		assertThat((List<Object>) target.get("objectList")).containsExactly(t800AsMap, gunAsMap);
	}

	@Test // DATAES-530
	public void readGenericList() {

		Document source = Document.create();
		source.put("objectList", Arrays.asList(t800AsMap, gunAsMap));

		Skynet target = mappingElasticsearchConverter.read(Skynet.class, source);

		assertThat(target.getObjectList()).containsExactly(t800, gun);
	}

	@Test // DATAES-530
	public void genericWriteListWithList() {

		Skynet skynet = new Skynet();
		skynet.objectList = new ArrayList<>();
		skynet.objectList.add(Arrays.asList(t800, gun));

		Map<String, Object> target = writeToMap(skynet);

		assertThat((List<Object>) target.get("objectList")).containsExactly(Arrays.asList(t800AsMap, gunAsMap));
	}

	@Test // DATAES-530
	public void readGenericListList() {

		Document source = Document.create();
		source.put("objectList", Collections.singletonList(Arrays.asList(t800AsMap, gunAsMap)));

		Skynet target = mappingElasticsearchConverter.read(Skynet.class, source);

		assertThat(target.getObjectList()).containsExactly(Arrays.asList(t800, gun));
	}

	@Test // DATAES-530
	public void writeGenericMap() {

		Skynet skynet = new Skynet();
		skynet.objectMap = new LinkedHashMap<>();
		skynet.objectMap.put("gun", gun);
		skynet.objectMap.put("grenade", grenade);

		Map<String, Object> target = writeToMap(skynet);

		assertThat((Map<String, Object>) target.get("objectMap")).containsEntry("gun", gunAsMap).containsEntry("grenade",
				grenadeAsMap);
	}

	@Test // DATAES-530
	public void readGenericMap() {

		Document source = Document.create();
		source.put("objectMap", Collections.singletonMap("glock19", gunAsMap));

		Skynet target = mappingElasticsearchConverter.read(Skynet.class, source);

		assertThat(target.getObjectMap()).containsEntry("glock19", gun);
	}

	@Test // DATAES-530
	public void writeGenericMapMap() {

		Skynet skynet = new Skynet();
		skynet.objectMap = new LinkedHashMap<>();
		skynet.objectMap.put("inventory", Collections.singletonMap("glock19", gun));

		Map<String, Object> target = writeToMap(skynet);

		assertThat((Map<String, Object>) target.get("objectMap")).containsEntry("inventory",
				Collections.singletonMap("glock19", gunAsMap));
	}

	@Test // DATAES-530
	public void readGenericMapMap() {

		Document source = Document.create();
		source.put("objectMap", Collections.singletonMap("inventory", Collections.singletonMap("glock19", gunAsMap)));

		Skynet target = mappingElasticsearchConverter.read(Skynet.class, source);

		assertThat(target.getObjectMap()).containsEntry("inventory", Collections.singletonMap("glock19", gun));
	}

	@Test // DATAES-530
	public void readsNestedEntity() {

		sarahAsMap.put("address", gratiotAveAsMap);

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.getAddress()).isEqualTo(observatoryRoad);
	}

	@Test // DATAES-530
	public void readsNestedObjectEntity() {

		Document source = Document.create();
		source.put("object", t800AsMap);

		Skynet target = mappingElasticsearchConverter.read(Skynet.class, source);

		assertThat(target.getObject()).isEqualTo(t800);
	}

	@Test // DATAES-530
	public void writesAliased() {
		assertThat(writeToMap(rifle)).containsEntry("_class", "rifle").doesNotContainValue(Rifle.class.getName());
	}

	@Test // DATAES-530
	public void writesNestedAliased() {

		t800.inventoryList = Collections.singletonList(rifle);
		Map<String, Object> target = writeToMap(t800);

		assertThat((List) target.get("inventoryList")).contains(rifleAsMap);
	}

	@Test // DATAES-530
	public void readsAliased() {
		assertThat(mappingElasticsearchConverter.read(Inventory.class, rifleAsMap)).isEqualTo(rifle);
	}

	@Test // DATAES-530
	public void readsNestedAliased() {

		t800AsMap.put("inventoryList", Collections.singletonList(rifleAsMap));

		assertThat(mappingElasticsearchConverter.read(Person.class, t800AsMap).getInventoryList()).containsExactly(rifle);
	}

	@Test // DATAES-530
	public void appliesCustomConverterForWrite() {
		assertThat(writeToMap(shotGun)).isEqualTo(shotGunAsMap);
	}

	@Test // DATAES-530
	public void appliesCustomConverterForRead() {
		assertThat(mappingElasticsearchConverter.read(Inventory.class, shotGunAsMap)).isEqualTo(shotGun);
	}

	@Test // DATAES-530
	public void writeSubTypeCorrectly() {

		sarahConnor.address = bigBunsCafe;

		Map<String, Object> target = writeToMap(sarahConnor);

		assertThat(target.get("address")).isEqualTo(bigBunsCafeAsMap);
	}

	@Test // DATAES-530
	public void readSubTypeCorrectly() {

		sarahAsMap.put("address", bigBunsCafeAsMap);

		Person target = mappingElasticsearchConverter.read(Person.class, sarahAsMap);

		assertThat(target.address).isEqualTo(bigBunsCafe);
	}

	@Test // DATAES-716
	void shouldWriteLocalDate() throws JSONException {
		Person person = new Person();
		person.setId("4711");
		person.setFirstName("John");
		person.setLastName("Doe");
		person.birthDate = LocalDate.of(2000, 8, 22);
		person.gender = Gender.MAN;

		String expected = '{' + //
				"  \"id\": \"4711\"," + //
				"  \"first-name\": \"John\"," + //
				"  \"last-name\": \"Doe\"," + //
				"  \"birth-date\": \"22.08.2000\"," + //
				"  \"gender\": \"MAN\"" + //
				'}';
		Document document = Document.create();
		mappingElasticsearchConverter.write(person, document);
		String json = document.toJson();

		assertEquals(expected, json, false);
	}

	@Test // DATAES-924
	@DisplayName("should write list of LocalDate")
	void shouldWriteListOfLocalDate() throws JSONException {

		LocalDatesEntity entity = new LocalDatesEntity();
		entity.setId("4711");
		entity.setDates(Arrays.asList(LocalDate.of(2020, 9, 15), LocalDate.of(2019, 5, 1)));
		String expected = "{\n" + //
				"  \"id\": \"4711\",\n" + //
				"  \"dates\": [\"15.09.2020\", \"01.05.2019\"]\n" + //
				"}\n"; //

		Document document = Document.create();
		mappingElasticsearchConverter.write(entity, document);
		String json = document.toJson();

		assertEquals(expected, json, false);
	}

	@Test // DATAES-716
	void shouldReadLocalDate() {
		Document document = Document.create();
		document.put("id", "4711");
		document.put("first-name", "John");
		document.put("last-name", "Doe");
		document.put("birth-date", "22.08.2000");
		document.put("gender", "MAN");

		Person person = mappingElasticsearchConverter.read(Person.class, document);

		assertThat(person.getId()).isEqualTo("4711");
		assertThat(person.getBirthDate()).isEqualTo(LocalDate.of(2000, 8, 22));
		assertThat(person.getGender()).isEqualTo(Gender.MAN);
	}

	@Test // DATAES-924
	@DisplayName("should read list of LocalDate")
	void shouldReadListOfLocalDate() {

		Document document = Document.create();
		document.put("id", "4711");
		document.put("dates", new String[] { "15.09.2020", "01.05.2019" });

		LocalDatesEntity entity = mappingElasticsearchConverter.read(LocalDatesEntity.class, document);

		assertThat(entity.getId()).isEqualTo("4711");
		assertThat(entity.getDates()).hasSize(2).containsExactly(LocalDate.of(2020, 9, 15), LocalDate.of(2019, 5, 1));
	}

	@Test // DATAES-763
	void writeEntityWithMapDataType() {

		Notification notification = new Notification();
		notification.setFromEmail("from@email.com");
		notification.setToEmail("to@email.com");
		Map<String, Object> data = new HashMap<>();
		data.put("documentType", "abc");
		data.put("content", null);
		notification.params = data;
		notification.id = 1L;

		Document document = Document.create();
		mappingElasticsearchConverter.write(notification, document);
		assertThat(document).isEqualTo(notificationAsMap);
	}

	@Test // DATAES-763
	void readEntityWithMapDataType() {

		Document document = Document.create();
		document.put("id", 1L);
		document.put("fromEmail", "from@email.com");
		document.put("toEmail", "to@email.com");
		Map<String, Object> data = new HashMap<>();
		data.put("documentType", "abc");
		data.put("content", null);
		document.put("params", data);

		Notification notification = mappingElasticsearchConverter.read(Notification.class, document);
		assertThat(notification.params.get("documentType")).isEqualTo("abc");
		assertThat(notification.params.get("content")).isNull();
	}

	@Test // DATAES-795
	void readGenericMapWithSimpleTypes() {
		Map<String, Object> mapWithSimpleValues = new HashMap<>();
		mapWithSimpleValues.put("int", 1);
		mapWithSimpleValues.put("string", "string");
		mapWithSimpleValues.put("boolean", true);

		Document document = Document.create();
		document.put("schemaLessObject", mapWithSimpleValues);

		SchemaLessObjectWrapper wrapper = mappingElasticsearchConverter.read(SchemaLessObjectWrapper.class, document);
		assertThat(wrapper.getSchemaLessObject()).isEqualTo(mapWithSimpleValues);
	}

	@Test // DATAES-797
	void readGenericListWithMaps() {
		Map<String, Object> simpleMap = new HashMap<>();
		simpleMap.put("int", 1);

		List<Map<String, Object>> listWithSimpleMap = new ArrayList<>();
		listWithSimpleMap.add(simpleMap);

		Map<String, List<Map<String, Object>>> mapWithSimpleList = new HashMap<>();
		mapWithSimpleList.put("someKey", listWithSimpleMap);

		Document document = Document.create();
		document.put("schemaLessObject", mapWithSimpleList);

		SchemaLessObjectWrapper wrapper = mappingElasticsearchConverter.read(SchemaLessObjectWrapper.class, document);
		assertThat(wrapper.getSchemaLessObject()).isEqualTo(mapWithSimpleList);
	}

	@Test // DATAES-799
	void shouldNotWriteSeqNoPrimaryTermProperty() {
		EntityWithSeqNoPrimaryTerm entity = new EntityWithSeqNoPrimaryTerm();
		entity.seqNoPrimaryTerm = new SeqNoPrimaryTerm(1L, 2L);
		Document document = Document.create();

		mappingElasticsearchConverter.write(entity, document);

		assertThat(document).doesNotContainKey("seqNoPrimaryTerm");
	}

	@Test // DATAES-799
	void shouldNotReadSeqNoPrimaryTermProperty() {
		Document document = Document.create().append("seqNoPrimaryTerm", emptyMap());

		EntityWithSeqNoPrimaryTerm entity = mappingElasticsearchConverter.read(EntityWithSeqNoPrimaryTerm.class, document);

		assertThat(entity.seqNoPrimaryTerm).isNull();
	}

	@Test // DATAES-845
	void shouldWriteCollectionsWithNullValues() throws JSONException {
		EntityWithListProperty entity = new EntityWithListProperty();
		entity.setId("42");
		entity.setValues(Arrays.asList(null, "two", null, "four"));

		String expected = '{' + //
				"  \"id\": \"42\"," + //
				"  \"values\": [null, \"two\", null, \"four\"]" + //
				'}';
		Document document = Document.create();
		mappingElasticsearchConverter.write(entity, document);
		String json = document.toJson();

		assertEquals(expected, json, false);
	}

	@Test // DATAES-857
	void shouldWriteEntityWithListOfGeoPoints() throws JSONException {

		GeoPointListEntity entity = new GeoPointListEntity();
		entity.setId("42");
		List<GeoPoint> locations = Arrays.asList(new GeoPoint(12.34, 23.45), new GeoPoint(34.56, 45.67));
		entity.setLocations(locations);

		String expected = "{\n" + //
				"  \"id\": \"42\",\n" + //
				"  \"locations\": [\n" + //
				"    {\n" + //
				"      \"lat\": 12.34,\n" + //
				"      \"lon\": 23.45\n" + //
				"    },\n" + //
				"    {\n" + //
				"      \"lat\": 34.56,\n" + //
				"      \"lon\": 45.67\n" + //
				"    }\n" + //
				"  ]\n" + //
				"}"; //
		Document document = Document.create();

		mappingElasticsearchConverter.write(entity, document);
		String json = document.toJson();

		assertEquals(expected, json, false);
	}

	@Test // DATAES-857
	void shouldReadEntityWithListOfGeoPoints() {

		String json = "{\n" + //
				"  \"id\": \"42\",\n" + //
				"  \"locations\": [\n" + //
				"    {\n" + //
				"      \"lat\": 12.34,\n" + //
				"      \"lon\": 23.45\n" + //
				"    },\n" + //
				"    {\n" + //
				"      \"lat\": 34.56,\n" + //
				"      \"lon\": 45.67\n" + //
				"    }\n" + //
				"  ]\n" + //
				"}"; //

		Document document = Document.parse(json);

		GeoPointListEntity entity = mappingElasticsearchConverter.read(GeoPointListEntity.class, document);

		assertThat(entity.id).isEqualTo("42");
		assertThat(entity.locations).containsExactly(new GeoPoint(12.34, 23.45), new GeoPoint(34.56, 45.67));
	}

	@Test // DATAES-865
	void shouldWriteEntityWithMapAsObject() throws JSONException {

		Map<String, Object> map = new LinkedHashMap<>();
		map.put("foo", "bar");

		EntityWithObject entity = new EntityWithObject();
		entity.setId("42");
		entity.setContent(map);

		String expected = "{\n" + //
				"  \"id\": \"42\",\n" + //
				"  \"content\": {\n" + //
				"    \"foo\": \"bar\"\n" + //
				"  }\n" + //
				"}\n"; //

		Document document = Document.create();

		mappingElasticsearchConverter.write(entity, document);

		assertEquals(expected, document.toJson(), false);
	}

	@Test // DATAES-920
	@DisplayName("should write null value if configured")
	void shouldWriteNullValueIfConfigured() throws JSONException {

		EntityWithNullField entity = new EntityWithNullField();
		entity.setId("42");

		String expected = "{\n" + //
				"  \"id\": \"42\",\n" + //
				"  \"saved\": null\n" + //
				"}\n"; //

		Document document = Document.create();

		mappingElasticsearchConverter.write(entity, document);

		assertEquals(expected, document.toJson(), false);
	}

	@Nested
	class RangeTests {

		static final String JSON = "{"
				+ "\"_class\":\"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$RangeTests$RangeEntity\","
				+ "\"integerRange\":{\"gt\":\"1\",\"lt\":\"10\"}," //
				+ "\"floatRange\":{\"gte\":\"1.2\",\"lte\":\"2.5\"}," //
				+ "\"longRange\":{\"gt\":\"2\",\"lte\":\"5\"}," //
				+ "\"doubleRange\":{\"gte\":\"3.2\",\"lt\":\"7.4\"}," //
				+ "\"dateRange\":{\"gte\":\"1970-01-01T00:00:00.000Z\",\"lte\":\"1970-01-01T01:00:00.000Z\"}," //
				+ "\"localDateRange\":{\"gte\":\"2021-07-06\"}," //
				+ "\"localTimeRange\":{\"gte\":\"00:30:00.000\",\"lt\":\"02:30:00.000\"}," //
				+ "\"localDateTimeRange\":{\"gt\":\"2021-01-01T00:30:00.000\",\"lt\":\"2021-01-01T02:30:00.000\"}," //
				+ "\"offsetTimeRange\":{\"gte\":\"00:30:00.000+02:00\",\"lt\":\"02:30:00.000+02:00\"}," //
				+ "\"zonedDateTimeRange\":{\"gte\":\"2021-01-01T00:30:00.000+02:00\",\"lte\":\"2021-01-01T00:30:00.000+02:00\"}," //
				+ "\"nullRange\":null}";

		@Test
		public void shouldReadRanges() throws JSONException {

			// given
			Document source = Document.parse(JSON);

			// when
			RangeEntity entity = mappingElasticsearchConverter.read(RangeEntity.class, source);

			// then
			assertThat(entity) //
					.isNotNull() //
					.satisfies(e -> {
						assertThat(e.getIntegerRange()).isEqualTo(Range.open(1, 10));
						assertThat(e.getFloatRange()).isEqualTo(Range.closed(1.2f, 2.5f));
						assertThat(e.getLongRange()).isEqualTo(Range.leftOpen(2l, 5l));
						assertThat(e.getDoubleRange()).isEqualTo(Range.rightOpen(3.2d, 7.4d));
						assertThat(e.getDateRange()).isEqualTo(Range.closed(new Date(0), new Date(60 * 60 * 1000)));
						assertThat(e.getLocalDateRange())
								.isEqualTo(Range.rightUnbounded(Range.Bound.inclusive(LocalDate.of(2021, 7, 6))));
						assertThat(e.getLocalTimeRange()).isEqualTo(Range.rightOpen(LocalTime.of(0, 30), LocalTime.of(2, 30)));
						assertThat(e.getLocalDateTimeRange())
								.isEqualTo(Range.open(LocalDateTime.of(2021, 1, 1, 0, 30), LocalDateTime.of(2021, 1, 1, 2, 30)));
						assertThat(e.getOffsetTimeRange())
								.isEqualTo(Range.rightOpen(OffsetTime.of(LocalTime.of(0, 30), ZoneOffset.ofHours(2)),
										OffsetTime.of(LocalTime.of(2, 30), ZoneOffset.ofHours(2))));
						assertThat(e.getZonedDateTimeRange()).isEqualTo(
								Range.just(ZonedDateTime.of(LocalDate.of(2021, 1, 1), LocalTime.of(0, 30), ZoneOffset.ofHours(2))));
						assertThat(e.getNullRange()).isNull();
					});
		}

		@Test
		public void shouldWriteRanges() throws JSONException {

			// given
			Document source = Document.parse(JSON);
			RangeEntity entity = new RangeEntity();
			entity.setIntegerRange(Range.open(1, 10));
			entity.setFloatRange(Range.closed(1.2f, 2.5f));
			entity.setLongRange(Range.leftOpen(2l, 5l));
			entity.setDoubleRange(Range.rightOpen(3.2d, 7.4d));
			entity.setDateRange(Range.closed(new Date(0), new Date(60 * 60 * 1000)));
			entity.setLocalDateRange(Range.rightUnbounded(Range.Bound.inclusive(LocalDate.of(2021, 7, 6))));
			entity.setLocalTimeRange(Range.rightOpen(LocalTime.of(0, 30), LocalTime.of(2, 30)));
			entity
					.setLocalDateTimeRange(Range.open(LocalDateTime.of(2021, 1, 1, 0, 30), LocalDateTime.of(2021, 1, 1, 2, 30)));
			entity.setOffsetTimeRange(Range.rightOpen(OffsetTime.of(LocalTime.of(0, 30), ZoneOffset.ofHours(2)),
					OffsetTime.of(LocalTime.of(2, 30), ZoneOffset.ofHours(2))));
			entity.setZonedDateTimeRange(
					Range.just(ZonedDateTime.of(LocalDate.of(2021, 1, 1), LocalTime.of(0, 30), ZoneOffset.ofHours(2))));
			entity.setNullRange(null);

			// when
			Document document = mappingElasticsearchConverter.mapObject(entity);

			// then
			assertThat(document).isEqualTo(source);
		}

		@org.springframework.data.elasticsearch.annotations.Document(indexName = "test-index-range-entity-mapper")
		class RangeEntity {

			@Id private String id;
			@Field(type = FieldType.Integer_Range) private Range<Integer> integerRange;
			@Field(type = FieldType.Float_Range) private Range<Float> floatRange;
			@Field(type = FieldType.Long_Range) private Range<Long> longRange;
			@Field(type = FieldType.Double_Range) private Range<Double> doubleRange;
			@Field(type = FieldType.Date_Range) private Range<Date> dateRange;
			@Field(type = FieldType.Date_Range, format = DateFormat.year_month_day) private Range<LocalDate> localDateRange;
			@Field(type = FieldType.Date_Range,
					format = DateFormat.hour_minute_second_millis) private Range<LocalTime> localTimeRange;
			@Field(type = FieldType.Date_Range,
					format = DateFormat.date_hour_minute_second_millis) private Range<LocalDateTime> localDateTimeRange;
			@Field(type = FieldType.Date_Range, format = DateFormat.time) private Range<OffsetTime> offsetTimeRange;
			@Field(type = FieldType.Date_Range) private Range<ZonedDateTime> zonedDateTimeRange;
			@Field(type = FieldType.Date_Range, storeNullValue = true) private Range<ZonedDateTime> nullRange;

			public String getId() {
				return id;
			}

			public Range<Integer> getIntegerRange() {
				return integerRange;
			}

			public Range<Float> getFloatRange() {
				return floatRange;
			}

			public Range<Long> getLongRange() {
				return longRange;
			}

			public Range<Double> getDoubleRange() {
				return doubleRange;
			}

			public Range<Date> getDateRange() {
				return dateRange;
			}

			public Range<LocalDate> getLocalDateRange() {
				return localDateRange;
			}

			public Range<LocalTime> getLocalTimeRange() {
				return localTimeRange;
			}

			public Range<LocalDateTime> getLocalDateTimeRange() {
				return localDateTimeRange;
			}

			public Range<OffsetTime> getOffsetTimeRange() {
				return offsetTimeRange;
			}

			public Range<ZonedDateTime> getZonedDateTimeRange() {
				return zonedDateTimeRange;
			}

			public Range<ZonedDateTime> getNullRange() {
				return nullRange;
			}

			public void setId(String id) {
				this.id = id;
			}

			public void setIntegerRange(Range<Integer> integerRange) {
				this.integerRange = integerRange;
			}

			public void setFloatRange(Range<Float> floatRange) {
				this.floatRange = floatRange;
			}

			public void setLongRange(Range<Long> longRange) {
				this.longRange = longRange;
			}

			public void setDoubleRange(Range<Double> doubleRange) {
				this.doubleRange = doubleRange;
			}

			public void setDateRange(Range<Date> dateRange) {
				this.dateRange = dateRange;
			}

			public void setLocalDateRange(Range<LocalDate> localDateRange) {
				this.localDateRange = localDateRange;
			}

			public void setLocalTimeRange(Range<LocalTime> localTimeRange) {
				this.localTimeRange = localTimeRange;
			}

			public void setLocalDateTimeRange(Range<LocalDateTime> localDateTimeRange) {
				this.localDateTimeRange = localDateTimeRange;
			}

			public void setOffsetTimeRange(Range<OffsetTime> offsetTimeRange) {
				this.offsetTimeRange = offsetTimeRange;
			}

			public void setZonedDateTimeRange(Range<ZonedDateTime> zonedDateTimeRange) {
				this.zonedDateTimeRange = zonedDateTimeRange;
			}

			public void setNullRange(Range<ZonedDateTime> nullRange) {
				this.nullRange = nullRange;
			}

		}
	}

	@Nested
	class GeoJsonUnitTests {
		private GeoJsonEntity entity;

		@BeforeEach
		void setup() {
			GeoJsonMultiLineString multiLineString = GeoJsonMultiLineString.of(Arrays.asList( //
					GeoJsonLineString.of(new Point(12, 34), new Point(56, 78)), //
					GeoJsonLineString.of(new Point(90, 12), new Point(34, 56)) //
			));
			GeoJsonPolygon geoJsonPolygon = GeoJsonPolygon //
					.of(new Point(12, 34), new Point(56, 78), new Point(90, 12), new Point(12, 34)) //
					.withInnerRing(new Point(21, 43), new Point(65, 87), new Point(9, 21), new Point(21, 43));
			GeoJsonMultiPolygon geoJsonMultiPolygon = GeoJsonMultiPolygon.of(
					Arrays.asList(GeoJsonPolygon.of(new Point(12, 34), new Point(56, 78), new Point(90, 12), new Point(12, 34)),
							GeoJsonPolygon.of(new Point(21, 43), new Point(65, 87), new Point(9, 21), new Point(21, 43))));
			GeoJsonPoint geoJsonPoint = GeoJsonPoint.of(12, 34);
			GeoJsonGeometryCollection geoJsonGeometryCollection = GeoJsonGeometryCollection
					.of(Arrays.asList(GeoJsonPoint.of(12, 34), GeoJsonPolygon
							.of(GeoJsonLineString.of(new Point(12, 34), new Point(56, 78), new Point(90, 12), new Point(12, 34)))));

			entity = new GeoJsonEntity();
			entity.setId("42");
			entity.setPoint1(GeoJsonPoint.of(12, 34));
			entity.setPoint2(GeoJsonPoint.of(56, 78));
			entity.setMultiPoint1(GeoJsonMultiPoint.of(new Point(12, 34), new Point(56, 78), new Point(90, 12)));
			entity.setMultiPoint2(GeoJsonMultiPoint.of(new Point(90, 12), new Point(56, 78), new Point(12, 34)));
			entity.setLineString1(GeoJsonLineString.of(new Point(12, 34), new Point(56, 78), new Point(90, 12)));
			entity.setLineString2(GeoJsonLineString.of(new Point(90, 12), new Point(56, 78), new Point(12, 34)));
			entity.setMultiLineString1(multiLineString);
			entity.setMultiLineString2(multiLineString);
			entity.setPolygon1(geoJsonPolygon);
			entity.setPolygon2(geoJsonPolygon);
			entity.setMultiPolygon1(geoJsonMultiPolygon);
			entity.setMultiPolygon2(geoJsonMultiPolygon);
			entity.setGeometryCollection1(geoJsonGeometryCollection);
			entity.setGeometryCollection2(geoJsonGeometryCollection);
		}

		@Test // DATAES-930
		@DisplayName("should write GeoJson properties")
		void shouldWriteGeoJsonProperties() throws JSONException {

			String json = "{\n" + //
					"  \"id\": \"42\",\n" + //
					"  \"point1\": {\n" + //
					"    \"type\": \"Point\",\n" + //
					"    \"coordinates\": [12.0, 34.0]\n" + //
					"  },\n" + //
					"  \"point2\": {\n" + //
					"    \"type\": \"Point\",\n" + //
					"    \"coordinates\": [56.0, 78.0]\n" + //
					"  },\n" + //
					"  \"multiPoint1\": {\n" + //
					"    \"type\": \"MultiPoint\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [12.0, 34.0],\n" + //
					"      [56.0, 78.0],\n" + //
					"      [90.0, 12.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPoint2\": {\n" + //
					"    \"type\": \"MultiPoint\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [90.0, 12.0],\n" + //
					"      [56.0, 78.0],\n" + //
					"      [12.0, 34.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"lineString1\": {\n" + //
					"    \"type\": \"LineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [12.0, 34.0],\n" + //
					"      [56.0, 78.0],\n" + //
					"      [90.0, 12.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"lineString2\": {\n" + //
					"    \"type\": \"LineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [90.0, 12.0],\n" + //
					"      [56.0, 78.0],\n" + //
					"      [12.0, 34.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiLineString1\":{\n" + //
					"    \"type\": \"MultiLineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12.0, 34.0], [56.0, 78.0]],\n" + //
					"      [[90.0, 12.0], [34.0, 56.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiLineString2\":{\n" + //
					"    \"type\": \"MultiLineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12.0, 34.0], [56.0, 78.0]],\n" + //
					"      [[90.0, 12.0], [34.0, 56.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"polygon1\":{\n" + //
					"    \"type\": \"Polygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12.0, 34.0],[56.0, 78.0],[90.0, 12.0],[12.0, 34.0]],\n" + //
					"      [[21.0, 43.0],[65.0, 87.0],[9.0, 21.0],[21.0, 43.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"polygon2\":{\n" + //
					"    \"type\": \"Polygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12.0, 34.0],[56.0, 78.0],[90.0, 12.0],[12.0, 34.0]],\n" + //
					"      [[21.0, 43.0],[65.0, 87.0],[9.0, 21.0],[21.0, 43.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPolygon1\":{\n" + //
					"    \"type\": \"MultiPolygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[[12.0, 34.0],[56.0, 78.0],[90.0, 12.0],[12.0, 34.0]]],\n" + //
					"      [[[21.0, 43.0],[65.0, 87.0],[9.0, 21.0],[21.0, 43.0]]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPolygon2\":{\n" + //
					"    \"type\": \"MultiPolygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[[12.0, 34.0],[56.0, 78.0],[90.0, 12.0],[12.0, 34.0]]],\n" + //
					"      [[[21.0, 43.0],[65.0, 87.0],[9.0, 21.0],[21.0, 43.0]]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"geometryCollection1\": {\n" + //
					"    \"type\": \"GeometryCollection\",\n" + //
					"    \"geometries\": [\n" + //
					"      {\n" + //
					"        \"type\": \"Point\",\n" + //
					"        \"coordinates\": [12.0, 34.0]\n" + //
					"      },\n" + //
					"      {\n" + //
					"        \"type\": \"Polygon\",\n" + //
					"        \"coordinates\": [\n" + //
					"          [[12.0, 34.0], [56.0, 78.0], [90.0, 12.0], [12.0, 34.0]]\n" + //
					"        ]\n" + //
					"      }\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"geometryCollection2\": {\n" + //
					"    \"type\": \"GeometryCollection\",\n" + //
					"    \"geometries\": [\n" + //
					"      {\n" + //
					"        \"type\": \"Point\",\n" + //
					"        \"coordinates\": [12.0, 34.0]\n" + //
					"      },\n" + //
					"      {\n" + //
					"        \"type\": \"Polygon\",\n" + //
					"        \"coordinates\": [\n" + //
					"          [[12.0, 34.0], [56.0, 78.0], [90.0, 12.0], [12.0, 34.0]]\n" + //
					"        ]\n" + //
					"      }\n" + //
					"    ]\n" + //
					"  }\n" + //
					"}\n"; //

			Document document = Document.create();

			mappingElasticsearchConverter.write(entity, document);

			assertEquals(json, document.toJson(), false);
		}

		@Test // DATAES-930
		@DisplayName("should read GeoJson properties")
		void shouldReadGeoJsonProperties() {

			// make sure we can read int values as well
			String json = "{\n" + //
					"  \"id\": \"42\",\n" + //
					"  \"point1\": {\n" + //
					"    \"type\": \"Point\",\n" + //
					"    \"coordinates\": [12, 34]\n" + //
					"  },\n" + //
					"  \"point2\": {\n" + //
					"    \"type\": \"Point\",\n" + //
					"    \"coordinates\": [56, 78]\n" + //
					"  },\n" + //
					"  \"multiPoint1\": {\n" + //
					"    \"type\": \"MultiPoint\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [12.0, 34],\n" + //
					"      [56, 78.0],\n" + //
					"      [90, 12.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPoint2\": {\n" + //
					"    \"type\": \"MultiPoint\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [90, 12.0],\n" + //
					"      [56, 78.0],\n" + //
					"      [12.0, 34]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"lineString1\": {\n" + //
					"    \"type\": \"LineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [12.0, 34],\n" + //
					"      [56, 78.0],\n" + //
					"      [90, 12.0]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"lineString2\": {\n" + //
					"    \"type\": \"LineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [90, 12.0],\n" + //
					"      [56, 78.0],\n" + //
					"      [12.0, 34]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiLineString1\":{\n" + //
					"    \"type\": \"MultiLineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12, 34.0], [56, 78.0]],\n" + //
					"      [[90.0, 12], [34.0, 56]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiLineString2\":{\n" + //
					"    \"type\": \"MultiLineString\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12.0, 34], [56.0, 78]],\n" + //
					"      [[90, 12.0], [34, 56.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"polygon1\":{\n" + //
					"    \"type\": \"Polygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12, 34.0],[56.0, 78],[90, 12.0],[12.0, 34]],\n" + //
					"      [[21.0, 43],[65, 87.0],[9.0, 21],[21, 43.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"polygon2\":{\n" + //
					"    \"type\": \"Polygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[12, 34.0],[56.0, 78],[90, 12.0],[12.0, 34]],\n" + //
					"      [[21.0, 43],[65, 87.0],[9.0, 21],[21, 43.0]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPolygon1\":{\n" + //
					"    \"type\": \"MultiPolygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[[12, 34.0],[56.0, 78],[90, 12.0],[12.0, 34]]],\n" + //
					"      [[[21.0, 43],[65, 87.0],[9.0, 21],[21, 43.0]]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"multiPolygon2\":{\n" + //
					"    \"type\": \"MultiPolygon\",\n" + //
					"    \"coordinates\": [\n" + //
					"      [[[12, 34.0],[56.0, 78],[90, 12.0],[12.0, 34]]],\n" + //
					"      [[[21.0, 43],[65, 87.0],[9.0, 21],[21, 43.0]]]\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"geometryCollection1\": {\n" + //
					"    \"type\": \"GeometryCollection\",\n" + //
					"    \"geometries\": [\n" + //
					"      {\n" + //
					"        \"type\": \"Point\",\n" + //
					"        \"coordinates\": [12, 34.0]\n" + //
					"      },\n" + //
					"      {\n" + //
					"        \"type\": \"Polygon\",\n" + //
					"        \"coordinates\": [\n" + //
					"          [[12.0, 34], [56, 78.0], [90.0, 12], [12, 34.0]]\n" + //
					"        ]\n" + //
					"      }\n" + //
					"    ]\n" + //
					"  },\n" + //
					"  \"geometryCollection2\": {\n" + //
					"    \"type\": \"GeometryCollection\",\n" + //
					"    \"geometries\": [\n" + //
					"      {\n" + //
					"        \"type\": \"Point\",\n" + //
					"        \"coordinates\": [12, 34.0]\n" + //
					"      },\n" + //
					"      {\n" + //
					"        \"type\": \"Polygon\",\n" + //
					"        \"coordinates\": [\n" + //
					"          [[12.0, 34], [56, 78.0], [90.0, 12], [12, 34.0]]\n" + //
					"        ]\n" + //
					"      }\n" + //
					"    ]\n" + //
					"  }\n" + //
					"}\n"; //

			GeoJsonEntity mapped = mappingElasticsearchConverter.read(GeoJsonEntity.class, Document.parse(json));

			assertThat(entity).isEqualTo(mapped);
		}
	}

	@Test // #1454
	@DisplayName("should write type hints if configured")
	void shouldWriteTypeHintsIfConfigured() throws JSONException {

		((SimpleElasticsearchMappingContext) mappingElasticsearchConverter.getMappingContext()).setWriteTypeHints(true);
		PersonWithCars person = new PersonWithCars();
		person.setId("42");
		person.setName("Smith");
		Car car1 = new Car();
		car1.setModel("Ford Mustang");
		Car car2 = new ElectricCar();
		car2.setModel("Porsche Taycan");
		person.setCars(Arrays.asList(car1, car2));

		String expected = "{\n" + //
				"  \"_class\": \"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$PersonWithCars\",\n"
				+ "  \"id\": \"42\",\n" + //
				"  \"name\": \"Smith\",\n" + //
				"  \"cars\": [\n" + //
				"    {\n" + //
				"      \"model\": \"Ford Mustang\"\n" + //
				"    },\n" + //
				"    {\n" + //
				"      \"_class\": \"org.springframework.data.elasticsearch.core.convert.MappingElasticsearchConverterUnitTests$ElectricCar\",\n"
				+ "      \"model\": \"Porsche Taycan\"\n" + //
				"    }\n" + //
				"  ]\n" + //
				"}\n"; //

		Document document = Document.create();

		mappingElasticsearchConverter.write(person, document);

		assertEquals(expected, document.toJson(), true);
	}

	@Test // #1454
	@DisplayName("should not write type hints if configured")
	void shouldNotWriteTypeHintsIfNotConfigured() throws JSONException {

		((SimpleElasticsearchMappingContext) mappingElasticsearchConverter.getMappingContext()).setWriteTypeHints(false);
		PersonWithCars person = new PersonWithCars();
		person.setId("42");
		person.setName("Smith");
		Car car1 = new Car();
		car1.setModel("Ford Mustang");
		Car car2 = new ElectricCar();
		car2.setModel("Porsche Taycan");
		person.setCars(Arrays.asList(car1, car2));

		String expected = "{\n" + //
				"  \"id\": \"42\",\n" + //
				"  \"name\": \"Smith\",\n" + //
				"  \"cars\": [\n" + //
				"    {\n" + //
				"      \"model\": \"Ford Mustang\"\n" + //
				"    },\n" + //
				"    {\n" + //
				"      \"model\": \"Porsche Taycan\"\n" + //
				"    }\n" + //
				"  ]\n" + //
				"}\n"; //

		Document document = Document.create();

		mappingElasticsearchConverter.write(person, document);

		assertEquals(expected, document.toJson(), true);
	}

	private Map<String, Object> writeToMap(Object source) {

		Document sink = Document.create();
		mappingElasticsearchConverter.write(source, sink);
		return sink;
	}

	// region entities
	public static class Sample {
		@Nullable public @ReadOnlyProperty String readOnly;
		@Nullable public @Transient String annotatedTransientProperty;
		@Nullable public transient String javaTransientProperty;
		@Nullable public String property;

		@Nullable
		public String getReadOnly() {
			return readOnly;
		}

		public void setReadOnly(@Nullable String readOnly) {
			this.readOnly = readOnly;
		}

		@Nullable
		public String getAnnotatedTransientProperty() {
			return annotatedTransientProperty;
		}

		public void setAnnotatedTransientProperty(@Nullable String annotatedTransientProperty) {
			this.annotatedTransientProperty = annotatedTransientProperty;
		}

		@Nullable
		public String getJavaTransientProperty() {
			return javaTransientProperty;
		}

		public void setJavaTransientProperty(@Nullable String javaTransientProperty) {
			this.javaTransientProperty = javaTransientProperty;
		}

		@Nullable
		public String getProperty() {
			return property;
		}

		public void setProperty(@Nullable String property) {
			this.property = property;
		}
	}

	static class Person {
		@Nullable @Id String id;
		@Nullable String name;
		@Nullable @Field(name = "first-name") String firstName;
		@Nullable @Field(name = "last-name") String lastName;
		@Nullable @Field(name = "birth-date", type = FieldType.Date, format = {},
				pattern = "dd.MM.uuuu") LocalDate birthDate;
		@Nullable Gender gender;
		@Nullable Address address;
		@Nullable List<Person> coWorkers;
		@Nullable List<Inventory> inventoryList;
		@Nullable Map<String, Address> shippingAddresses;
		@Nullable Map<String, Inventory> inventoryMap;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public String getName() {
			return name;
		}

		public void setName(@Nullable String name) {
			this.name = name;
		}

		@Nullable
		public String getFirstName() {
			return firstName;
		}

		public void setFirstName(@Nullable String firstName) {
			this.firstName = firstName;
		}

		@Nullable
		public String getLastName() {
			return lastName;
		}

		public void setLastName(@Nullable String lastName) {
			this.lastName = lastName;
		}

		@Nullable
		public LocalDate getBirthDate() {
			return birthDate;
		}

		public void setBirthDate(@Nullable LocalDate birthDate) {
			this.birthDate = birthDate;
		}

		@Nullable
		public Gender getGender() {
			return gender;
		}

		public void setGender(@Nullable Gender gender) {
			this.gender = gender;
		}

		@Nullable
		public Address getAddress() {
			return address;
		}

		public void setAddress(@Nullable Address address) {
			this.address = address;
		}

		@Nullable
		public List<Person> getCoWorkers() {
			return coWorkers;
		}

		public void setCoWorkers(@Nullable List<Person> coWorkers) {
			this.coWorkers = coWorkers;
		}

		@Nullable
		public List<Inventory> getInventoryList() {
			return inventoryList;
		}

		public void setInventoryList(@Nullable List<Inventory> inventoryList) {
			this.inventoryList = inventoryList;
		}

		@Nullable
		public Map<String, Address> getShippingAddresses() {
			return shippingAddresses;
		}

		public void setShippingAddresses(@Nullable Map<String, Address> shippingAddresses) {
			this.shippingAddresses = shippingAddresses;
		}

		@Nullable
		public Map<String, Inventory> getInventoryMap() {
			return inventoryMap;
		}

		public void setInventoryMap(@Nullable Map<String, Inventory> inventoryMap) {
			this.inventoryMap = inventoryMap;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;

			Person person = (Person) o;

			if (id != null ? !id.equals(person.id) : person.id != null)
				return false;
			if (name != null ? !name.equals(person.name) : person.name != null)
				return false;
			if (firstName != null ? !firstName.equals(person.firstName) : person.firstName != null)
				return false;
			if (lastName != null ? !lastName.equals(person.lastName) : person.lastName != null)
				return false;
			if (birthDate != null ? !birthDate.equals(person.birthDate) : person.birthDate != null)
				return false;
			if (gender != person.gender)
				return false;
			if (address != null ? !address.equals(person.address) : person.address != null)
				return false;
			if (coWorkers != null ? !coWorkers.equals(person.coWorkers) : person.coWorkers != null)
				return false;
			if (inventoryList != null ? !inventoryList.equals(person.inventoryList) : person.inventoryList != null)
				return false;
			if (shippingAddresses != null ? !shippingAddresses.equals(person.shippingAddresses)
					: person.shippingAddresses != null)
				return false;
			return inventoryMap != null ? inventoryMap.equals(person.inventoryMap) : person.inventoryMap == null;
		}

		@Override
		public int hashCode() {
			int result = id != null ? id.hashCode() : 0;
			result = 31 * result + (name != null ? name.hashCode() : 0);
			result = 31 * result + (firstName != null ? firstName.hashCode() : 0);
			result = 31 * result + (lastName != null ? lastName.hashCode() : 0);
			result = 31 * result + (birthDate != null ? birthDate.hashCode() : 0);
			result = 31 * result + (gender != null ? gender.hashCode() : 0);
			result = 31 * result + (address != null ? address.hashCode() : 0);
			result = 31 * result + (coWorkers != null ? coWorkers.hashCode() : 0);
			result = 31 * result + (inventoryList != null ? inventoryList.hashCode() : 0);
			result = 31 * result + (shippingAddresses != null ? shippingAddresses.hashCode() : 0);
			result = 31 * result + (inventoryMap != null ? inventoryMap.hashCode() : 0);
			return result;
		}
	}

	static class LocalDatesEntity {
		@Nullable @Id private String id;
		@Nullable @Field(name = "dates", type = FieldType.Date, format = DateFormat.custom,
				pattern = "dd.MM.uuuu") private List<LocalDate> dates;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public List<LocalDate> getDates() {
			return dates;
		}

		public void setDates(@Nullable List<LocalDate> dates) {
			this.dates = dates;
		}
	}

	enum Gender {

		MAN("1"), MACHINE("0");

		String theValue;

		Gender(String theValue) {
			this.theValue = theValue;
		}

		public String getTheValue() {
			return theValue;
		}
	}

	interface Inventory {

		String getLabel();
	}

	static class Gun implements Inventory {
		final String label;
		final int shotsPerMagazine;

		public Gun(@Nullable String label, int shotsPerMagazine) {
			this.label = label;
			this.shotsPerMagazine = shotsPerMagazine;
		}

		@Override
		public String getLabel() {
			return label;
		}

		public int getShotsPerMagazine() {
			return shotsPerMagazine;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (o == null || getClass() != o.getClass())
				return false;

			Gun gun = (Gun) o;

			if (shotsPerMagazine != gun.shotsPerMagazine)
				return false;
			return label.equals(gun.label);
		}

		@Override
		public int hashCode() {
			int result = label.hashCode();
			result = 31 * result + shotsPerMagazine;
			return result;
		}
	}

	static class Grenade implements Inventory {
		final String label;

		public Grenade(String label) {
			this.label = label;
		}

		@Override
		public String getLabel() {
			return label;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (!(o instanceof Grenade))
				return false;

			Grenade grenade = (Grenade) o;

			return label.equals(grenade.label);
		}

		@Override
		public int hashCode() {
			return label.hashCode();
		}
	}

	@TypeAlias("rifle")
	static class Rifle implements Inventory {

		final String label;
		final double weight;
		final int maxShotsPerMagazine;

		public Rifle(String label, double weight, int maxShotsPerMagazine) {
			this.label = label;
			this.weight = weight;
			this.maxShotsPerMagazine = maxShotsPerMagazine;
		}

		@Override
		public String getLabel() {
			return label;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (!(o instanceof Rifle))
				return false;

			Rifle rifle = (Rifle) o;

			if (Double.compare(rifle.weight, weight) != 0)
				return false;
			if (maxShotsPerMagazine != rifle.maxShotsPerMagazine)
				return false;
			return label.equals(rifle.label);
		}

		@Override
		public int hashCode() {
			int result;
			long temp;
			result = label.hashCode();
			temp = Double.doubleToLongBits(weight);
			result = 31 * result + (int) (temp ^ (temp >>> 32));
			result = 31 * result + maxShotsPerMagazine;
			return result;
		}
	}

	static class ShotGun implements Inventory {

		private final String label;

		public ShotGun(String label) {
			this.label = label;
		}

		@Override
		public String getLabel() {
			return label;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (!(o instanceof ShotGun))
				return false;

			ShotGun shotGun = (ShotGun) o;

			return label.equals(shotGun.label);
		}

		@Override
		public int hashCode() {
			return label.hashCode();
		}
	}

	static class Address {
		@Nullable private Point location;
		@Nullable private String street;
		@Nullable private String city;

		@Nullable
		public Point getLocation() {
			return location;
		}

		public void setLocation(@Nullable Point location) {
			this.location = location;
		}

		@Nullable
		public String getStreet() {
			return street;
		}

		public void setStreet(@Nullable String street) {
			this.street = street;
		}

		@Nullable
		public String getCity() {
			return city;
		}

		public void setCity(@Nullable String city) {
			this.city = city;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (!(o instanceof Address))
				return false;

			Address address = (Address) o;

			if (location != null ? !location.equals(address.location) : address.location != null)
				return false;
			if (street != null ? !street.equals(address.street) : address.street != null)
				return false;
			return city != null ? city.equals(address.city) : address.city == null;
		}

		@Override
		public int hashCode() {
			int result = location != null ? location.hashCode() : 0;
			result = 31 * result + (street != null ? street.hashCode() : 0);
			result = 31 * result + (city != null ? city.hashCode() : 0);
			return result;
		}
	}

	static class Place extends Address {
		@Nullable private String name;

		@Nullable
		public String getName() {
			return name;
		}

		public void setName(@Nullable String name) {
			this.name = name;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o)
				return true;
			if (!(o instanceof Place))
				return false;

			Place place = (Place) o;

			return name != null ? name.equals(place.name) : place.name == null;
		}

		@Override
		public int hashCode() {
			return name != null ? name.hashCode() : 0;
		}
	}

	static class Skynet {
		@Nullable private Object object;
		@Nullable private List<Object> objectList;
		@Nullable private Map<String, Object> objectMap;

		@Nullable
		public Object getObject() {
			return object;
		}

		public void setObject(@Nullable Object object) {
			this.object = object;
		}

		@Nullable
		public List<Object> getObjectList() {
			return objectList;
		}

		public void setObjectList(@Nullable List<Object> objectList) {
			this.objectList = objectList;
		}

		@Nullable
		public Map<String, Object> getObjectMap() {
			return objectMap;
		}

		public void setObjectMap(@Nullable Map<String, Object> objectMap) {
			this.objectMap = objectMap;
		}
	}

	static class Notification {
		@Nullable private Long id;
		@Nullable private String fromEmail;
		@Nullable private String toEmail;
		@Nullable private Map<String, Object> params;

		@Nullable
		public Long getId() {
			return id;
		}

		public void setId(@Nullable Long id) {
			this.id = id;
		}

		@Nullable
		public String getFromEmail() {
			return fromEmail;
		}

		public void setFromEmail(@Nullable String fromEmail) {
			this.fromEmail = fromEmail;
		}

		@Nullable
		public String getToEmail() {
			return toEmail;
		}

		public void setToEmail(@Nullable String toEmail) {
			this.toEmail = toEmail;
		}

		@Nullable
		public Map<String, Object> getParams() {
			return params;
		}

		public void setParams(@Nullable Map<String, Object> params) {
			this.params = params;
		}
	}

	@WritingConverter
	static class ShotGunToMapConverter implements Converter<ShotGun, Map<String, Object>> {

		@Override
		public Map<String, Object> convert(ShotGun source) {

			LinkedHashMap<String, Object> target = new LinkedHashMap<>();
			target.put("model", source.getLabel());
			target.put("_class", ShotGun.class.getName());
			return target;
		}
	}

	@ReadingConverter
	static class MapToShotGunConverter implements Converter<Map<String, Object>, ShotGun> {

		@Override
		public ShotGun convert(Map<String, Object> source) {
			return new ShotGun(source.get("model").toString());
		}
	}

	static class Car {
		@Nullable private String name;
		@Nullable private String model;

		@Nullable
		public String getName() {
			return name;
		}

		public void setName(@Nullable String name) {
			this.name = name;
		}

		@Nullable
		public String getModel() {
			return model;
		}

		public void setModel(@Nullable String model) {
			this.model = model;
		}
	}

	@org.springframework.data.elasticsearch.annotations.Document(indexName = "test-index-geo-core-entity-mapper")
	static class GeoEntity {
		@Nullable @Id private String id;
		// geo shape - Spring Data
		@Nullable private Box box;
		@Nullable private Circle circle;
		@Nullable private Polygon polygon;
		// geo point - Custom implementation + Spring Data
		@Nullable @GeoPointField private Point pointA;
		@Nullable private GeoPoint pointB;
		@Nullable @GeoPointField private String pointC;
		@Nullable @GeoPointField private double[] pointD;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public Box getBox() {
			return box;
		}

		public void setBox(@Nullable Box box) {
			this.box = box;
		}

		@Nullable
		public Circle getCircle() {
			return circle;
		}

		public void setCircle(@Nullable Circle circle) {
			this.circle = circle;
		}

		@Nullable
		public Polygon getPolygon() {
			return polygon;
		}

		public void setPolygon(@Nullable Polygon polygon) {
			this.polygon = polygon;
		}

		@Nullable
		public Point getPointA() {
			return pointA;
		}

		public void setPointA(@Nullable Point pointA) {
			this.pointA = pointA;
		}

		@Nullable
		public GeoPoint getPointB() {
			return pointB;
		}

		public void setPointB(@Nullable GeoPoint pointB) {
			this.pointB = pointB;
		}

		@Nullable
		public String getPointC() {
			return pointC;
		}

		public void setPointC(@Nullable String pointC) {
			this.pointC = pointC;
		}

		@Nullable
		public double[] getPointD() {
			return pointD;
		}

		public void setPointD(@Nullable double[] pointD) {
			this.pointD = pointD;
		}
	}

	static class SchemaLessObjectWrapper {
		@Nullable private Map<String, Object> schemaLessObject;

		@Nullable
		public Map<String, Object> getSchemaLessObject() {
			return schemaLessObject;
		}

		public void setSchemaLessObject(@Nullable Map<String, Object> schemaLessObject) {
			this.schemaLessObject = schemaLessObject;
		}
	}

	@org.springframework.data.elasticsearch.annotations.Document(
			indexName = "test-index-entity-with-seq-no-primary-term-mapper")
	static class EntityWithSeqNoPrimaryTerm {
		@Nullable private SeqNoPrimaryTerm seqNoPrimaryTerm;

		@Nullable
		public SeqNoPrimaryTerm getSeqNoPrimaryTerm() {
			return seqNoPrimaryTerm;
		}

		public void setSeqNoPrimaryTerm(@Nullable SeqNoPrimaryTerm seqNoPrimaryTerm) {
			this.seqNoPrimaryTerm = seqNoPrimaryTerm;
		}
	}

	static class EntityWithListProperty {
		@Nullable @Id private String id;
		@Nullable private List<String> values;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public List<String> getValues() {
			return values;
		}

		public void setValues(@Nullable List<String> values) {
			this.values = values;
		}
	}

	static class GeoPointListEntity {
		@Nullable private @Id String id;
		@Nullable private List<GeoPoint> locations;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public List<GeoPoint> getLocations() {
			return locations;
		}

		public void setLocations(@Nullable List<GeoPoint> locations) {
			this.locations = locations;
		}
	}

	static class EntityWithObject {
		@Nullable @Id private String id;
		@Nullable private Object content;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public Object getContent() {
			return content;
		}

		public void setContent(@Nullable Object content) {
			this.content = content;
		}
	}

	static class EntityWithNullField {
		@Nullable @Id private String id;
		@Nullable @Field(type = FieldType.Text) private String notSaved;
		@Nullable @Field(type = FieldType.Text, storeNullValue = true) private String saved;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public String getNotSaved() {
			return notSaved;
		}

		public void setNotSaved(@Nullable String notSaved) {
			this.notSaved = notSaved;
		}

		@Nullable
		public String getSaved() {
			return saved;
		}

		public void setSaved(@Nullable String saved) {
			this.saved = saved;
		}
	}

	private static class ElectricCar extends Car {}

	private static class PersonWithCars {
		@Id @Nullable String id;
		@Field(type = FieldType.Text) @Nullable private String name;
		@Field(type = FieldType.Nested) @Nullable private List<? extends Car> cars;

		@Nullable
		public String getId() {
			return id;
		}

		public void setId(@Nullable String id) {
			this.id = id;
		}

		@Nullable
		public String getName() {
			return name;
		}

		public void setName(@Nullable String name) {
			this.name = name;
		}

		@Nullable
		public List<? extends Car> getCars() {
			return cars;
		}

		public void setCars(@Nullable List<Car> cars) {
			this.cars = cars;
		}
	}
	// endregion
}
